/* TTMPSockPlat.c
 *
 * TTMPSocket implementation for Linux.
 *
 * Copyright (C) 2005-2006 TomTom BV <http://www.tomtom.com/>
 * Author 0: Dimitry Andric <dimitry.andric@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2
 * as published by the Free Software Foundation.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "TTMPSock.h"
#include "TTMPPriv.h"

int TTMPSocketAddressParse(cTTMPAddress addr, TTMPSocketAddress *sa)
{
	size_t len;
	char *s;
	int port;
	char buf[TTMP_MAX_ADDRESS + 1];
	in_addr_t ia;

	if (addr == NULL || (len = strlen(addr)) > TTMP_MAX_ADDRESS)
		return TTMP_ERR_ARGUMENT;

	memcpy(buf, addr, len + 1);

	/* Check for a port number */
	port = -1;
	s = strchr(buf, ':');
	if (s != NULL) {
		port = atoi(s + 1);
		if (port == 0)
			return TTMP_ERR_ARGUMENT;
		*s = '\0';
		len = s - buf;
	}

	ia = inet_addr(buf);
	if (ia == INADDR_NONE) {
		/* It isn't an IPv4 address, assume it's a socket name. */
		if (port != -1) {
			/* Don't accept port numbers with sockets. */
			return TTMP_ERR_ARGUMENT;
		}
		sa->size = sizeof(struct sockaddr_un);
		sa->domain = AF_UNIX;
		sa->addr.sun.sun_family = AF_UNIX;
		memcpy(sa->addr.sun.sun_path, buf, len + 1);
	} else {
		/* It's an IPv4 address. */
		if (port == -1) {
			/* If no port was specified, use a hardcoded default. */
			port = 0x5454;
		} else if (port < 0 || port > 65535) {
			/* Don't accept illegal port numbers. */
			return TTMP_ERR_ARGUMENT;
		}
		sa->size = sizeof(struct sockaddr_in);
		sa->domain = AF_INET;
		sa->addr.sin.sin_family = AF_INET;
		sa->addr.sin.sin_port = htons(port);
		sa->addr.sin.sin_addr.s_addr = ia;
	}

	return TTMP_ERR_NONE;
}

int TTMPSocketNonBlocking(TTMPSocket sock)
{
	int res;

	/* Set socket to nonblocking */
	res = fcntl(sock, F_GETFL, 0);
	if (res < 0 || fcntl(sock, F_SETFL, res | O_NONBLOCK) < 0)
		return TTMP_ERR_SYSTEM;

	/* Set close-on-exec flag */
	res = fcntl(sock, F_SETFD, FD_CLOEXEC);
	if (res < 0)
		return TTMP_ERR_SYSTEM;

	return TTMP_ERR_NONE;
}

int TTMPSocketConnect(TTMPSocket sock, const TTMPSocketAddress *sa, TInt32 timeout)
{
	fd_set wset;
	int res;
	int err;
	socklen_t optlen;

	if (SOCKET_INVALID(sock) || sa == NULL || sa->domain == INVALID_DOMAIN)
		return TTMP_ERR_ARGUMENT;

	if (connect(sock, (struct sockaddr *) &sa->addr, sa->size) != SOCKET_ERROR)
		return TTMP_ERR_NONE; /* Immediately connected */

	if (errno != EINPROGRESS) {
		TTMPDBG("connect immediately failed, errno %d\n", errno);
		return TTMP_ERR_SYSTEM; /* Something else went wrong */
	}

	/* Still busy connecting, wait for completion */
	FD_ZERO(&wset);
	FD_SET(sock, &wset);
	res = TTMPSocketSelect(sock, NULL, &wset, NULL, timeout);
	if (res != TTMP_ERR_NONE)
		return res;

	if (!FD_ISSET(sock, &wset)) {
		/* Shouldn't happen */
		TTMPERR("socket not set in write fdset\n");
		return TTMP_ERR_INTERNAL;
	}

	optlen = sizeof err;
	res = getsockopt(sock, SOL_SOCKET, SO_ERROR, (void *) &err, &optlen);
	if (res == SOCKET_ERROR) {
		TTMPERR("getsockopt failed, errno %d\n", errno);
		return TTMP_ERR_INTERNAL;
	}
	if (err != 0) {
		TTMPDBG("connect failed later, errno %d\n", err);
		return TTMP_ERR_SYSTEM;
	}

	/* Connected successfully */
	return TTMP_ERR_NONE;
}

/* EOF */
