/* TTMP.h
 *
 * TomTom Message Passing Interface declarations.
 *
 * Copyright (C) 2005-2006 TomTom BV <http://www.tomtom.com/>
 * Author 0: Serhiy Tkachenko <Serhiy.Tkachenko@tomtom.com>
 * Author 1: Dimitry Andric <dimitry.andric@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2
 * as published by the Free Software Foundation.
 */

#ifndef INCLUDED_TTMP_H
#define INCLUDED_TTMP_H

#include <stdio.h>
#include "basetypes.h" /* OSAL types */

#ifdef __cplusplus
extern "C" {
#endif

/* Forward declarations */
typedef struct sTTMPServer     *TTMPServer;     /**< \brief Server object */
typedef struct sTTMPConnection *TTMPConnection; /**< \brief Connection object */
typedef struct sTTMPMessage    *TTMPMessage;    /**< \brief Message object that can be sent via a connection */
/* Const versions */
typedef const struct sTTMPServer     *cTTMPServer;
typedef const struct sTTMPConnection *cTTMPConnection;
typedef const struct sTTMPMessage    *cTTMPMessage;


/** \brief Error codes definition. Negative values indicate errors. */
#define TTMP_ERR_NONE     0       /**< \brief Last library call succeeded */
#define TTMP_ERR_INTERNAL -1001   /**< \brief Unknown internal error */
#define TTMP_ERR_SVRNULL  -1002   /**< \brief NULL server passed */
#define TTMP_ERR_SVRINVAL -1003   /**< \brief Invalid server passed */
#define TTMP_ERR_CONNULL  -1004   /**< \brief NULL connection passed */
#define TTMP_ERR_CONINVAL -1005   /**< \brief Invalid connection passed */
#define TTMP_ERR_MSGNULL  -1006   /**< \brief NULL message passed */
#define TTMP_ERR_MSGINVAL -1007   /**< \brief Invalid message passed */
#define TTMP_ERR_ARGUMENT -1008   /**< \brief Invalid other argument */
#define TTMP_ERR_NOMEMORY -1009   /**< \brief Out of memory */
#define TTMP_ERR_TIMEOUT  -1010   /**< \brief Timeout elapsed */
#define TTMP_ERR_INCOMPAT -1011   /**< \brief Client and server version incompatible */
#define TTMP_ERR_NOSERVER -1012   /**< \brief No such server */
#define TTMP_ERR_NOCONN   -1013   /**< \brief No such connection */
#define TTMP_ERR_NOMSG    -1014   /**< \brief No such message */
#define TTMP_ERR_NOFIELD  -1015   /**< \brief No such field */
#define TTMP_ERR_OVERFLOW -1016   /**< \brief Caller buffer too small */
#define TTMP_ERR_NOTIMPL  -1017   /**< \brief Not implemented */
#define TTMP_ERR_SYSTEM   -1018   /**< \brief System call error */
#define TTMP_ERR_NOLISTEN -1019   /**< \brief Server not yet listening */
#define TTMP_ERR_NOTCONN  -1020   /**< \brief Not connected */
#define TTMP_ERR_MSGCORR  -1021   /**< \brief Messsage corrupt */
#define TTMP_ERR_PROTOVER -1022   /**< \brief Incompatible protocol version */
#define TTMP_ERR_DISCONN  -1023   /**< \brief Already connected, disconnect first */
#define TTMP_ERR_CONNEOF  -1024   /**< \brief Connection EOF (peer disconnected) */

/** \brief Connection timeout definition

Every blocking connection is associated with maximum time-out any of its call could
be blocked. If the timeout expires, corresponding call fails.
Timeout is measured in microseconds.
*/
#define TTMP_TO_NONE      0       /**< \brief No timeout (non-blocking) */
#define TTMP_TO_INFINITE  -1      /**< \brief Infinite timeout (blocking) */
#define TTMP_TO_DEFAULT   -2      /**< \brief Default timeout */

#define TTMP_TO_DEFVAL    10000   /**< \brief Default timeout value (microseconds) */

#define TTMP_MAX_ADDRESS  107     /**< \brief Maximum address length, without NUL terminator */
typedef char* TTMPAddress;        /**< \brief Connection address */
typedef const char* cTTMPAddress; /**< \brief Connection address (const) */
typedef TUint16 TUniChr;          /**< \brief Unicode character */


#define TTMP_MAX_FIELDNAME 32     /**< \brief Maximum message field name length, without NUL terminator */
typedef char TTMPFieldName[TTMP_MAX_FIELDNAME + 1];
typedef const char cTTMPFieldName[TTMP_MAX_FIELDNAME + 1];

#define UNUSED_PARAM(x) do { (void) (x); } while (0)


/* Error information */

/** \brief Gets error description

\param err  Error code
\param buf  Where to store the error description
\param size Size of buffer

Possible return values:
TTMP_ERR_NONE      Successful
TTMP_ERR_SYSTEM    System call error
TTMP_ERR_OVERFLOW  Caller buffer too small
*/

int TTMPErrorString(int err, char *buf, size_t size);


/* Message manipulation */

/** \brief Creates a message object

\param msg  Where to store the new message object

Possible return values:
TTMP_ERR_NONE      Successful
TTMP_ERR_ARGUMENT  Invalid argument(s)
TTMP_ERR_NOMEMORY  Out of memory
*/
int TTMPMessageCreate(TTMPMessage *msg);

/** \brief Destroys a message object

Releases all resources allocated to this message.

\param msg  Message object to destroy. If you pass a NULL value, or an invalid
message object, you will get an error.

Possible return values:
TTMP_ERR_NONE      Successful
TTMP_ERR_MSGNULL   NULL message passed
TTMP_ERR_MSGINVAL  Invalid message passed
*/
int TTMPMessageDestroy(TTMPMessage msg);

/** \brief Check if a message object is valid

Returns TTMP_ERR_NONE if the message object is valid
Returns an error if the message object is invalid

\param msg  Message object to validate.

Possible return values:
TTMP_ERR_NONE      Successful
TTMP_ERR_MSGNULL   NULL message passed
TTMP_ERR_MSGINVAL  Invalid message passed
*/
int TTMPMessageValidate(TTMPMessage msg);

/** \brief Adds a field to the message or changes the value of an existing one

\param msg   Message object to add field to
\param name  Name of message field to add or change
\param val   Value of message field

Possible return values:
TTMP_ERR_NONE      Successful
TTMP_ERR_MSGNULL   NULL message passed
TTMP_ERR_MSGINVAL  Invalid message passed
TTMP_ERR_ARGUMENT  Invalid argument(s)
TTMP_ERR_NOMEMORY  Out of memory
*/
int TTMPSetFieldInt8  (TTMPMessage msg, cTTMPFieldName name, TInt8   val);
int TTMPSetFieldUint8 (TTMPMessage msg, cTTMPFieldName name, TUint8  val);
int TTMPSetFieldInt16 (TTMPMessage msg, cTTMPFieldName name, TInt16  val);
int TTMPSetFieldUint16(TTMPMessage msg, cTTMPFieldName name, TUint16 val);
int TTMPSetFieldInt32 (TTMPMessage msg, cTTMPFieldName name, TInt32  val);
int TTMPSetFieldUint32(TTMPMessage msg, cTTMPFieldName name, TUint32 val);
int TTMPSetFieldReal32(TTMPMessage msg, cTTMPFieldName name, TReal32 val);
int TTMPSetFieldReal64(TTMPMessage msg, cTTMPFieldName name, TReal64 val);
int TTMPSetFieldReal  (TTMPMessage msg, cTTMPFieldName name, TReal   val);
int TTMPSetFieldBool  (TTMPMessage msg, cTTMPFieldName name, TBool   val);
int TTMPSetFieldChr   (TTMPMessage msg, cTTMPFieldName name, TChr    val);
int TTMPSetFieldUniChr(TTMPMessage msg, cTTMPFieldName name, TUniChr val);
int TTMPSetFieldLong64(TTMPMessage msg, cTTMPFieldName name, TLong64 val);
int TTMPSetFieldStr   (TTMPMessage msg, cTTMPFieldName name, const TChr    *val);
int TTMPSetFieldUniStr(TTMPMessage msg, cTTMPFieldName name, const TUniChr *val);
int TTMPSetFieldBinary(TTMPMessage msg, cTTMPFieldName name, const void    *val, TUint32 size);

/** \brief Deletes a field from a message

\param msg   Message object to delete field from
\param name  Name of message field to delete

Possible return values:
TTMP_ERR_NONE      Successful
TTMP_ERR_MSGNULL   NULL message passed
TTMP_ERR_MSGINVAL  Invalid message passed
TTMP_ERR_ARGUMENT  Invalid argument(s)
TTMP_ERR_NOFIELD   No such field
*/
int TTMPDelField(TTMPMessage msg, cTTMPFieldName name);

/** \brief Gets the size of a field

\param msg   Message object to get field size from
\param name  Name of message field to get size from

Possible return values:
TTMP_ERR_NONE      Successful
TTMP_ERR_MSGNULL   NULL message passed
TTMP_ERR_MSGINVAL  Invalid message passed
TTMP_ERR_ARGUMENT  Invalid argument(s)
TTMP_ERR_NOFIELD   No such field
*/
int TTMPGetFieldSize(cTTMPMessage msg, cTTMPFieldName name);

/** \brief Reads a elementary type field of a message

\param msg   Message object to read field from
\param name  Name of message field to read
\param val   Buffer to put field value in

Possible return values:
TTMP_ERR_NONE      Successful
TTMP_ERR_MSGNULL   NULL message passed
TTMP_ERR_MSGINVAL  Invalid message passed
TTMP_ERR_ARGUMENT  Invalid argument(s)
TTMP_ERR_NOFIELD   No such field
*/
int TTMPGetFieldInt8  (cTTMPMessage msg, cTTMPFieldName name, TInt8   *val);
int TTMPGetFieldUint8 (cTTMPMessage msg, cTTMPFieldName name, TUint8  *val);
int TTMPGetFieldInt16 (cTTMPMessage msg, cTTMPFieldName name, TInt16  *val);
int TTMPGetFieldUint16(cTTMPMessage msg, cTTMPFieldName name, TUint16 *val);
int TTMPGetFieldInt32 (cTTMPMessage msg, cTTMPFieldName name, TInt32  *val);
int TTMPGetFieldUint32(cTTMPMessage msg, cTTMPFieldName name, TUint32 *val);
int TTMPGetFieldReal32(cTTMPMessage msg, cTTMPFieldName name, TReal32 *val);
int TTMPGetFieldReal64(cTTMPMessage msg, cTTMPFieldName name, TReal64 *val);
int TTMPGetFieldReal  (cTTMPMessage msg, cTTMPFieldName name, TReal   *val);
int TTMPGetFieldBool  (cTTMPMessage msg, cTTMPFieldName name, TBool   *val);
int TTMPGetFieldChr   (cTTMPMessage msg, cTTMPFieldName name, TChr    *val);
int TTMPGetFieldUniChr(cTTMPMessage msg, cTTMPFieldName name, TUniChr *val);
int TTMPGetFieldLong64(cTTMPMessage msg, cTTMPFieldName name, TLong64 *val);

/** \brief Read arbitrary-sized fields

First get the field size using \a TTMPGetFieldSize, allocate your buffer, and
then pass it and its size to these functions.

\param msg   Message object to read field from
\param name  Name of message field to read
\param val   Buffer to put field value in
\param size  Size of buffer

Possible return values:
TTMP_ERR_NONE      Successful
TTMP_ERR_MSGNULL   NULL message passed
TTMP_ERR_MSGINVAL  Invalid message passed
TTMP_ERR_ARGUMENT  Invalid argument(s)
TTMP_ERR_NOFIELD   No such field
TTMP_ERR_OVERFLOW  Caller buffer too small
*/
int TTMPGetFieldStr   (cTTMPMessage msg, cTTMPFieldName name, TChr    *val, TUint32 size);
int TTMPGetFieldUniStr(cTTMPMessage msg, cTTMPFieldName name, TUniChr *val, TUint32 size);
int TTMPGetFieldBinary(cTTMPMessage msg, cTTMPFieldName name, void    *val, TUint32 size);

/** \brief Detach arbitrary-sized fields

Gets a direct pointer to the allocated buffer of a Str, UniStr or Binary field,
and resets the internal pointer to NULL.  This saves memory copies, BUT YOU
ALSO GET THE RESPONSIBILITY FOR FREEING THE BUFFER.

\param msg   Message object to read field from
\param name  Name of message field to read
\param val   Where to store the buffer pointer

Possible return values:
TTMP_ERR_NONE      Successful
TTMP_ERR_MSGNULL   NULL message passed
TTMP_ERR_MSGINVAL  Invalid message passed
TTMP_ERR_ARGUMENT  Invalid argument(s)
TTMP_ERR_NOFIELD   No such field
*/
int TTMPDetachFieldStr   (cTTMPMessage msg, cTTMPFieldName name, TChr    **val);
int TTMPDetachFieldUniStr(cTTMPMessage msg, cTTMPFieldName name, TUniChr **val);
int TTMPDetachFieldBinary(cTTMPMessage msg, cTTMPFieldName name, void    **val);

/** \brief Clear a message object

Clears the message content. Can be used for sending multiple messages using one
message object.

\param msg  Message object to clear

Possible return values:
TTMP_ERR_NONE      Successful
TTMP_ERR_MSGNULL   NULL message passed
TTMP_ERR_MSGINVAL  Invalid message passed
*/
int TTMPMessageClear(TTMPMessage msg);

/** \brief Dump a message object

Dump the message content, for debug purposes.

\param msg  Message object to dump
\param fp   Where to dump

Possible return values:
TTMP_ERR_NONE      Successful
TTMP_ERR_MSGNULL   NULL message passed
TTMP_ERR_MSGINVAL  Invalid message passed
TTMP_ERR_ARGUMENT  Invalid argument(s)
*/
int TTMPMessageDump(cTTMPMessage msg, FILE *fp);


/** \brief Server object

The server object is used by message servers.  Before it can be used, it must
be created with the \a TTMPServerCreate() function.  The server can be blocking
or non-blocking.  If it is non-blocking, no call to its methods will block.
This is particularly useful when developing multithreaded applications.

The user application is responsible for deleting server objects using the \a
TTMPServerDestroy() function when they're not needed anymore.
*/

/** \brief Creates a new server object

\param server   Where to store the new server object
\param timeout  Timeout value to use for this server
\param major    Major version of implemented protocol
\param minor    Minor version of implemented protocol

Protocol version is used to detect incompatibilities in used protocols between
client and server.

Possible return values:
TTMP_ERR_NONE      Successful
TTMP_ERR_SVRNULL   NULL server passed
TTMP_ERR_ARGUMENT  Invalid argument(s)
TTMP_ERR_NOMEMORY  Out of memory
TTMP_ERR_SYSTEM    System call error
*/
int TTMPServerCreate(TTMPServer *server, TInt32 timeout, TInt32 major, TInt32 minor);

/** \brief Destroys a server object

\param msg  Server object to destroy. If you pass a NULL value, or an invalid
            server object, you will get an error.

Stops listening on this server, and releases all its resources.  After this
call the server object becomes invalid.

Possible return values:
TTMP_ERR_NONE      Successful
TTMP_ERR_SVRNULL   NULL server passed
TTMP_ERR_SVRINVAL  Invalid server passed
TTMP_ERR_SYSTEM    System call error
*/
int TTMPServerDestroy(TTMPServer server);

/** \brief Check if a server object is valid

Returns TTMP_ERR_NONE if the server object is valid
Returns an error if the server object is invalid

\param msg  Server object to validate.

Possible return values:
TTMP_ERR_NONE      Successful
TTMP_ERR_SVRNULL   NULL server passed
TTMP_ERR_SVRINVAL  Invalid server passed
*/
int TTMPServerValidate(TTMPServer svr);

/** \brief Starts listening on a given address

\param server   Server object created by \a TTMPServerCreate()
\param addr     Address to listen on

Possible return values:
TTMP_ERR_NONE      Successful
TTMP_ERR_SVRNULL   NULL server passed
TTMP_ERR_SVRINVAL  Invalid server passed
TTMP_ERR_NOMEMORY  Out of memory
TTMP_ERR_SYSTEM    System call error
*/
int TTMPServerListen(TTMPServer server, cTTMPAddress addr);

/** \brief Waits for incoming connections, and upon connection, returns one.

\param server   Server object created by \a TTMPServerCreate()
\param newconn  Where to store the new connection object.
                If return value is 0, this object should be used for all
                subsequent communication with connected client.  If this
                function fails, it's not touched.  After an accepted connection,
                the original connection is available for an accept again.

Possible return values:
TTMP_ERR_NONE      Successful
TTMP_ERR_SVRNULL   NULL server passed
TTMP_ERR_SVRINVAL  Invalid server passed
TTMP_ERR_ARGUMENT  Invalid argument(s)
TTMP_ERR_NOLISTEN  Server not yet listening
TTMP_ERR_SYSTEM    System call error
TTMP_ERR_TIMEOUT   Timeout elapsed
TTMP_ERR_INTERNAL  Internal error
TTMP_ERR_NOMEMORY  Out of memory
TTMP_ERR_CONNEOF   Peer disconnected
TTMP_ERR_PROTOVER  Incompatible protocol version
TTMP_ERR_INCOMPAT  Client and server version incompatible
*/
int TTMPServerAccept(TTMPServer server, TTMPConnection* newconn);

/** \brief Closes a server object

Stops listening on this server.  After this call the server is available for
another listen call, optionally followed by accept calls.

Possible return values:
TTMP_ERR_NONE      Successful
TTMP_ERR_SVRNULL   NULL server passed
TTMP_ERR_SVRINVAL  Invalid server passed
TTMP_ERR_SYSTEM    System call error
*/
int TTMPServerClose(TTMPServer server);


/** \brief Connection object

This connection object should be used by clients to connect to a server.
Client connection is a structure with a number of functions that can operate with it.
Before connection could be used, it should be created with \a TTMPCreateClientConnection() function.
Connection could be blocking or non-blocking. If a connection is non-blocking,
no call to its methods will block. This is particularly useful when developing
mult-threaded applications
User application is responsible for deleting connections when it is not needed using its
\a iDestroy() function.

Established connection is a "real" connection that allows clients and servers to exchange messges.
Established connection is a structure with a number of functions that can operate with it.
It is created internally as a result of iAccept() it iConnect() calls.
Connection could be blocking or non-blocking. If a connection is non-blocking,
no call to its methods will block. This is particularly useful when developing
mult-threaded applications
User application is responsible for deleting connections when it is not needed using its
\a iDestroy() function.
*/

/** \brief Creates a new connection object

\param conn     Where to store the new connection object
\param timeout  Timeout value to use with this connection
\param major    Major version of implemented protocol
\param minor    Minor version of implemented protocol

Protocol version is used to detect incompatibilities in used protocols between
client and server.

Possible return values:
TTMP_ERR_NONE      Successful
TTMP_ERR_CONNULL   NULL connection passed
TTMP_ERR_ARGUMENT  Invalid argument(s)
TTMP_ERR_NOMEMORY  Out of memory
*/
int TTMPConnectionCreate(TTMPConnection *conn, TInt32 timeout, TInt32 major, TInt32 minor);

/** \brief Destroys the connection

Disconnects any peer, closes the connection and destroys it.  After this call
the connection object becomes invalid.  If you pass a NULL value, no operation
is performed.

Possible return values:
TTMP_ERR_NONE      Successful
TTMP_ERR_CONNULL   NULL connection passed
TTMP_ERR_CONINVAL  Invalid connection passed
TTMP_ERR_SYSTEM    System call error
*/
int TTMPConnectionDestroy(TTMPConnection conn);

/** \brief Check if a connection object is valid

Returns TTMP_ERR_NONE if the connection object is valid
Returns an error if the connection object is invalid

\param conn  Connection object to validate.

Possible return values:
TTMP_ERR_NONE      Successful
TTMP_ERR_CONNULL   NULL connection passed
TTMP_ERR_CONINVAL  Invalid connection passed
*/
int TTMPConnectionValidate(TTMPConnection conn);

/** \brief Tries to connect to a given address, blocking

\param conn     Connection created by \a TTMPConnectionCreate()
\param addr     Address to connect to

Possible return values:
TTMP_ERR_NONE      Successful
TTMP_ERR_CONNULL   NULL connection passed
TTMP_ERR_CONINVAL  Invalid connection passed
TTMP_ERR_ARGUMENT  Invalid argument(s)
TTMP_ERR_DISCONN   Already connected, disconnect first
TTMP_ERR_SYSTEM    System call error
TTMP_ERR_INTERNAL  Internal error
TTMP_ERR_PROTOVER  Incompatible protocol version
TTMP_ERR_INCOMPAT  Client and server version incompatible
*/
int TTMPConnectionConnect(TTMPConnection conn, cTTMPAddress addr);

/** \brief Tries to connect to a given address, using specified timeout

\param conn     Connection created by \a TTMPConnectionCreate()
\param addr     Address to connect to
\param timeout  Timeout value to use (can be TTMP_TO_INFINITE to block)

Possible return values:
TTMP_ERR_NONE      Successful
TTMP_ERR_CONNULL   NULL connection passed
TTMP_ERR_CONINVAL  Invalid connection passed
TTMP_ERR_ARGUMENT  Invalid argument(s)
TTMP_ERR_DISCONN   Already connected, disconnect first
TTMP_ERR_SYSTEM    System call error
TTMP_ERR_INTERNAL  Internal error
TTMP_ERR_TIMEOUT   Timeout elapsed
TTMP_ERR_PROTOVER  Incompatible protocol version
TTMP_ERR_INCOMPAT  Client and server version incompatible
*/
int TTMPConnectionConnectTimeout(TTMPConnection conn, cTTMPAddress addr, TInt32 timeout);

/** \brief Sends a message over an open connection, using specified timeout

\param conn     An open connection object
\param msg      A message initialized with \a TTMPMessageCreate(), and filled
                with one or more \a TTMPSetField*() calls.
\param timeout  Timeout value to use (can be TTMP_TO_INFINITE to block)

Possible return values:
TTMP_ERR_NONE      Successful
TTMP_ERR_CONNULL   NULL connection passed
TTMP_ERR_CONINVAL  Invalid connection passed
TTMP_ERR_MSGNULL   NULL message passed
TTMP_ERR_MSGINVAL  Invalid message passed
TTMP_ERR_NOTCONN   Not connected
TTMP_ERR_NOMEMORY  Out of memory
TTMP_ERR_SYSTEM    System call error
TTMP_ERR_TIMEOUT   Timeout elapsed
TTMP_ERR_CONNEOF   Peer disconnected
TTMP_ERR_INTERNAL  Internal error
*/
int TTMPConnectionSendTimeout(TTMPConnection conn, cTTMPMessage msg, TInt32 timeout);

/** \brief Sends a message over an open connection, using default timeout

\param conn     An open connection object
\param msg      A message initialized with \a TTMPMessageCreate(), and filled
                with one or more \a TTMPSetField*() calls.

Possible return values:
TTMP_ERR_NONE      Successful
TTMP_ERR_CONNULL   NULL connection passed
TTMP_ERR_CONINVAL  Invalid connection passed
TTMP_ERR_MSGNULL   NULL message passed
TTMP_ERR_MSGINVAL  Invalid message passed
TTMP_ERR_NOTCONN   Not connected
TTMP_ERR_NOMEMORY  Out of memory
TTMP_ERR_SYSTEM    System call error
TTMP_ERR_TIMEOUT   Timeout elapsed
TTMP_ERR_CONNEOF   Peer disconnected
TTMP_ERR_INTERNAL  Internal error
*/
int TTMPConnectionSend(TTMPConnection conn, cTTMPMessage msg);

/** \brief Sends a message over an open connection, blocking

\param conn  An open connection object
\param msg   A message initialized with \a TTMPMessageCreate(), and filled with
             one or more \a TTMPSetField*() calls.

Possible return values:
TTMP_ERR_NONE      Successful
TTMP_ERR_CONNULL   NULL connection passed
TTMP_ERR_CONINVAL  Invalid connection passed
TTMP_ERR_MSGNULL   NULL message passed
TTMP_ERR_MSGINVAL  Invalid message passed
TTMP_ERR_NOTCONN   Not connected
TTMP_ERR_NOMEMORY  Out of memory
TTMP_ERR_SYSTEM    System call error
TTMP_ERR_CONNEOF   Peer disconnected
TTMP_ERR_INTERNAL  Internal error
*/
int TTMPConnectionSendBlocking(TTMPConnection conn, cTTMPMessage msg);

/** \brief Receives a message from an open connection, using specified timeout

\param conn     An open connection object
\param msg      Where to store the received message
\param timeout  Timeout value to use (can be TTMP_TO_INFINITE to block)

Possible return values:
TTMP_ERR_NONE      Successful
TTMP_ERR_CONNULL   NULL connection passed
TTMP_ERR_CONINVAL  Invalid connection passed
TTMP_ERR_ARGUMENT  Invalid argument(s)
TTMP_ERR_NOTCONN   Not connected
TTMP_ERR_NOMEMORY  Out of memory
TTMP_ERR_SYSTEM    System call error
TTMP_ERR_TIMEOUT   Timeout elapsed
TTMP_ERR_CONNEOF   Peer disconnected
TTMP_ERR_MSGCORR   Message corrupt
TTMP_ERR_INTERNAL  Internal error
*/
int TTMPConnectionReceiveTimeout(TTMPConnection conn, TTMPMessage *msg, TInt32 timeout);

/** \brief Receives a message from an open connection, using default timeout

\param conn  An open connection object
\param msg   Where to store the received message

Possible return values:
TTMP_ERR_NONE      Successful
TTMP_ERR_CONNULL   NULL connection passed
TTMP_ERR_CONINVAL  Invalid connection passed
TTMP_ERR_ARGUMENT  Invalid argument(s)
TTMP_ERR_NOTCONN   Not connected
TTMP_ERR_NOMEMORY  Out of memory
TTMP_ERR_SYSTEM    System call error
TTMP_ERR_TIMEOUT   Timeout elapsed
TTMP_ERR_CONNEOF   Peer disconnected
TTMP_ERR_MSGCORR   Message corrupt
TTMP_ERR_INTERNAL  Internal error
*/
int TTMPConnectionReceive(TTMPConnection conn, TTMPMessage *msg);

/** \brief Receives a message from an open connection, blocking

\param conn  An open connection object
\param msg   Where to store the received message

Possible return values:
TTMP_ERR_NONE      Successful
TTMP_ERR_CONNULL   NULL connection passed
TTMP_ERR_CONINVAL  Invalid connection passed
TTMP_ERR_ARGUMENT  Invalid argument(s)
TTMP_ERR_NOTCONN   Not connected
TTMP_ERR_NOMEMORY  Out of memory
TTMP_ERR_SYSTEM    System call error
TTMP_ERR_CONNEOF   Peer disconnected
TTMP_ERR_MSGCORR   Message corrupt
TTMP_ERR_INTERNAL  Internal error
*/
int TTMPConnectionReceiveBlocking(TTMPConnection conn, TTMPMessage *msg);

/** \brief Receives a message from one of multiple open connections, using specified timeout

\param conn     An array of open connection objects
\param num      Number of connection objects in array
\param timeout  Timeout value to use (can be TTMP_TO_INFINITE to block)
\param msg      Where to store the received message
\param index    Index of connection that received the message

Possible return values:
XXX TODO
*/
int TTMPConnectionReceiveMultiTimeout(TTMPConnection conn[], TUint32 num, TInt32 timeout, TTMPMessage *msg, TUint32 *index);

/** \brief Receives a message from one of multiple open connections, blocking

\param conn     An array of open connection objects
\param num      Number of connection objects in array
\param msg      Where to store the received message
\param index    Index of connection that received the message

Possible return values:
XXX TODO
*/
int TTMPConnectionReceiveMultiBlocking(TTMPConnection conn[], TUint32 num, TTMPMessage *msg, TUint32 *index);

/** \brief Closes the connection

After this call connection is available for another connect call.

Possible return values:
TTMP_ERR_NONE      Successful
TTMP_ERR_CONNULL   NULL connection passed
TTMP_ERR_CONINVAL  Invalid connection passed
TTMP_ERR_SYSTEM    System call error
*/
int TTMPConnectionClose(TTMPConnection conn);


#ifdef __cplusplus
}
#endif

#endif /* INCLUDED_TTMP_H */

/* EOF */
