/* TTMPMsg.c
 *
 * TTMPMessage implementation.
 *
 * Copyright (C) 2005-2006 TomTom BV <http://www.tomtom.com/>
 * Author 0: Dimitry Andric <dimitry.andric@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2
 * as published by the Free Software Foundation.
 */

#include <stdlib.h>
#include <string.h>
#include "TTMPPriv.h"
#include "TTMPMsg.h"
#include "TTMPCRC.h"
#include "TTMPPlat.h"
#include "hexdump.h"

#ifdef DMALLOC
#include "dmalloc.h"
#endif

int TTMPMessageCreate(TTMPMessage *msg)
{
	TTMPMessage newmsg;

	if (msg == NULL)
		return TTMP_ERR_ARGUMENT;

	newmsg = (TTMPMessage) malloc(sizeof(struct sTTMPMessage));
	if (newmsg == NULL)
		return TTMP_ERR_NOMEMORY;

	/* Reset field count and array */
	newmsg->magic = TTMP_MSG_MAGIC;
	newmsg->numfields = 0;
	newmsg->maxfields = 0;
	newmsg->fields = NULL;

	/* Return message object to caller */
	*msg = newmsg;

	return TTMP_ERR_NONE;
}

static void TTMPFieldClear(TTMPField *field)
{
	switch (field->type) {
	case TTMP_FT_Str:
	case TTMP_FT_UniStr:
	case TTMP_FT_Binary:
		free(field->val.data);
		break;
	}
}

static void TTMPMessageClearInternal(TTMPMessage msg)
{
	TUint32 i;

	if (msg->numfields <= 0 || msg->fields == NULL)
		return;

	for (i = 0; i < msg->numfields; ++i)
		TTMPFieldClear(&msg->fields[i]);

	free(msg->fields);
}

int TTMPMessageDestroy(TTMPMessage msg)
{
	CHECK_OBJ(MSG, msg);

	/* Clean out all fields, if any */
	TTMPMessageClearInternal(msg);

	/* Zap magic to provoke errors */
	msg->magic = 0;

	/* Delete message object */
	free(msg);

	return TTMP_ERR_NONE;
}

int TTMPMessageValidate(TTMPMessage msg)
{
	CHECK_OBJ(MSG, msg);
	return TTMP_ERR_NONE;
}

int TTMPMessageClear(TTMPMessage msg)
{
	CHECK_OBJ(MSG, msg);

	/* Clean out all fields, if any */
	TTMPMessageClearInternal(msg);

	/* Reset field count and array */
	msg->numfields = 0;
	msg->maxfields = 0;
	msg->fields = NULL;

	return TTMP_ERR_NONE;
}

static int TTMPFieldFind(cTTMPMessage msg, cTTMPFieldName name, TUint32 *index)
{
	TUint32 i;

	CHECK_OBJ(MSG, msg);
	if (name == NULL || name[0] == '\0' || strlen(name) > TTMP_MAX_FIELDNAME || index == NULL)
		return TTMP_ERR_ARGUMENT;

	/* Check if the field exists */
	for (i = 0; i < msg->numfields; ++i) {
		if (strncmp(msg->fields[i].name, name, TTMP_MAX_FIELDNAME) == 0) {
			/* Found it */
			*index = i;
			return TTMP_ERR_NONE;
		}
	}

	return TTMP_ERR_NOFIELD;
}

static int TTMPFieldSetElem(TTMPMessage msg, cTTMPFieldName name, TUint8 type, TUint32 size, TUint32 *index)
{
	TUint32 i;
	int res = TTMPFieldFind(msg, name, &i);
	if (res == TTMP_ERR_NONE) {
		/* Found it, overwrite the contents */
		TTMPFieldClear(&msg->fields[i]);
	} else if (res == TTMP_ERR_NOFIELD) {
		/* Field wasn't found, add a new one */
		if (msg->numfields + 1 > msg->maxfields) {
			TUint32 newmax = msg->maxfields == 0 ? 4 : msg->maxfields * 2;
			TTMPField *newfields = realloc(msg->fields, newmax * sizeof(TTMPField));
			if (newfields == NULL)
				return TTMP_ERR_NOMEMORY;
			msg->maxfields = newmax;
			msg->fields = newfields;
		}
		i = msg->numfields++;
		strncpy(msg->fields[i].name, name, TTMP_MAX_FIELDNAME + 1);
	} else {
		/* Other error, pass it up */
		return res;
	}

	msg->fields[i].type = type;
	msg->fields[i].size = size;
	/* value itself is set by caller */
	*index = i;
	return TTMP_ERR_NONE;
}

#define DEF_SETFIELD(_Type) \
	int TTMPSetField##_Type(TTMPMessage msg, cTTMPFieldName name, T##_Type val) \
	{ \
		TUint32 i; \
		int res = TTMPFieldSetElem(msg, name, TTMP_FT_##_Type, TTMP_FS_##_Type, &i); \
		if (res != TTMP_ERR_NONE) \
			return res; \
		msg->fields[i].val._Type = val; \
		return TTMP_ERR_NONE; \
	}

DEF_SETFIELD(Int8)
DEF_SETFIELD(Uint8)
DEF_SETFIELD(Int16)
DEF_SETFIELD(Uint16)
DEF_SETFIELD(Int32)
DEF_SETFIELD(Uint32)
DEF_SETFIELD(Real32)
DEF_SETFIELD(Real64)
DEF_SETFIELD(Real)
DEF_SETFIELD(Bool)
DEF_SETFIELD(Chr)
DEF_SETFIELD(UniChr)
DEF_SETFIELD(Long64)

static int TTMPFieldSetBlob(TTMPMessage msg, cTTMPFieldName name, TUint8 type, const void *val, TUint32 size)
{
	void *blob;
	TUint32 i;
	int res;

	CHECK_OBJ(MSG, msg);
	if (name == NULL || name[0] == '\0' || strlen(name) > TTMP_MAX_FIELDNAME)
		return TTMP_ERR_ARGUMENT;
	/* val has already been checked */

	/* Allocate space for the blob, and copy it */
	blob = malloc(size);
	if (blob == NULL)
		return TTMP_ERR_NOMEMORY;

	memcpy(blob, val, size);
	res = TTMPFieldSetElem(msg, name, type, size, &i);
	if (res != TTMP_ERR_NONE) {
		/* Don't leak in case of errors */
		free(blob);
		return res;
	}

	/* Just a pointer assignment */
	msg->fields[i].val.data = blob;
	return TTMP_ERR_NONE;
}

int TTMPSetFieldStr(TTMPMessage msg, cTTMPFieldName name, const TChr *val)
{
	return val == NULL ?
		TTMP_ERR_ARGUMENT :
		TTMPFieldSetBlob(msg, name, TTMP_FT_Str, val, (TUint32) (strlen((const char *) val) + 1));
}

/* Since there's no good implementation of Unicode string length available yet,
 * we roll our own here... TODO: Remove when there's a shared implementation. */
static size_t unistrlen(const TUniChr *s)
{
	size_t len;
	for (len = 0; *s != 0; ++s, ++len)
		;
	return len;
}

int TTMPSetFieldUniStr(TTMPMessage msg, cTTMPFieldName name, const TUniChr *val)
{
	return val == NULL ?
		TTMP_ERR_ARGUMENT :
		TTMPFieldSetBlob(msg, name, TTMP_FT_UniStr, val, (TUint32) ((unistrlen(val) + 1) * sizeof(TUniChr)));
}

int TTMPSetFieldBinary(TTMPMessage msg, cTTMPFieldName name, const void *val, TUint32 size)
{
	return val == NULL || size <= 0 ?
		TTMP_ERR_ARGUMENT :
		TTMPFieldSetBlob(msg, name, TTMP_FT_Binary, val, size);
}

/** \brief Deletes a field from a message */
int TTMPDelField(TTMPMessage msg, cTTMPFieldName name)
{
	TUint32 i;
	int res = TTMPFieldFind(msg, name, &i);
	if (res != TTMP_ERR_NONE)
		return res;

	/* Found it, now clean out the field */
	TTMPFieldClear(&msg->fields[i]);
	/* Move other fields up, if any */
	if (msg->numfields > i)
		memmove(&msg->fields[i], &msg->fields[i + 1], (msg->numfields - i) * sizeof(TTMPField));
	msg->numfields--;

	return TTMP_ERR_NONE;
}

/** \brief Gets the size of a field */
int TTMPGetFieldSize(cTTMPMessage msg, cTTMPFieldName name)
{
	TUint32 i;
	int res = TTMPFieldFind(msg, name, &i);
	if (res != TTMP_ERR_NONE)
		return res;

	/* Found it, now return the size */
	return msg->fields[i].size;
}

#define DEF_GETFIELD(_Type) \
	int TTMPGetField##_Type(cTTMPMessage msg, cTTMPFieldName name, T##_Type *val) \
	{ \
		TUint32 i; \
		int res; \
		if (val == NULL) \
			return TTMP_ERR_ARGUMENT; \
		res = TTMPFieldFind(msg, name, &i); \
		if (res != TTMP_ERR_NONE) \
			return res; \
		*val = msg->fields[i].val._Type; \
		return TTMP_ERR_NONE; \
	}

DEF_GETFIELD(Int8)
DEF_GETFIELD(Uint8)
DEF_GETFIELD(Int16)
DEF_GETFIELD(Uint16)
DEF_GETFIELD(Int32)
DEF_GETFIELD(Uint32)
DEF_GETFIELD(Real32)
DEF_GETFIELD(Real64)
DEF_GETFIELD(Real)
DEF_GETFIELD(Bool)
DEF_GETFIELD(Chr)
DEF_GETFIELD(UniChr)
DEF_GETFIELD(Long64)

static int TTMPFieldGetBlob(cTTMPMessage msg, cTTMPFieldName name, void *val, TUint32 size)
{
	TUint32 i;
	int res;
	if (val == NULL)
		return TTMP_ERR_ARGUMENT;
	res = TTMPFieldFind(msg, name, &i);
	if (res != TTMP_ERR_NONE)
		return res;
	if (msg->fields[i].size > size)
		return TTMP_ERR_OVERFLOW;
	memcpy(val, msg->fields[i].val.data, msg->fields[i].size);
	return TTMP_ERR_NONE;
}

int TTMPGetFieldStr(cTTMPMessage msg, cTTMPFieldName name, TChr *val, TUint32 size)
{
	return TTMPFieldGetBlob(msg, name, val, size);
}

int TTMPGetFieldUniStr(cTTMPMessage msg, cTTMPFieldName name, TUniChr *val, TUint32 size)
{
	return TTMPFieldGetBlob(msg, name, val, size);
}

int TTMPGetFieldBinary(cTTMPMessage msg, cTTMPFieldName name, void *val, TUint32 size)
{
	return TTMPFieldGetBlob(msg, name, val, size);
}

static int TTMPFieldDetachBlob(cTTMPMessage msg, cTTMPFieldName name, void **val)
{
	TUint32 i;
	int res;
	if (val == NULL)
		return TTMP_ERR_ARGUMENT;
	res = TTMPFieldFind(msg, name, &i);
	if (res != TTMP_ERR_NONE)
		return res;
	*val = msg->fields[i].val.data;
	msg->fields[i].val.data = NULL;
	msg->fields[i].size = 0;
	return TTMP_ERR_NONE;
}

int TTMPDetachFieldStr(cTTMPMessage msg, cTTMPFieldName name, TChr **val)
{
	return TTMPFieldDetachBlob(msg, name, (void **) val);
}

int TTMPDetachFieldUniStr(cTTMPMessage msg, cTTMPFieldName name, TUniChr **val)
{
	return TTMPFieldDetachBlob(msg, name, (void **) val);
}

int TTMPDetachFieldBinary(cTTMPMessage msg, cTTMPFieldName name, void **val)
{
	return TTMPFieldDetachBlob(msg, name, val);
}

int TTMPMessageDump(cTTMPMessage msg, FILE *fp)
{
	TUint32 i;
	TTMPField *f;

	CHECK_OBJ(MSG, msg);
	if (fp == NULL)
		return TTMP_ERR_ARGUMENT;

	//fprintf(fp, "Message fields used: %d, max %d\n", msg->numfields, msg->maxfields);
	for (i = 0; i < msg->numfields; ++i) {
		f = &msg->fields[i];
		fprintf(fp, "%2u: name \"%.*s\", type ", i, TTMP_MAX_FIELDNAME, f->name);
		switch (f->type) {
		case TTMP_FT_Int8:
			fprintf(fp, "Int8, value %d\n", (int) f->val.Int8);
			break;
		case TTMP_FT_Uint8:
			fprintf(fp, "Uint8, value %u\n", (unsigned) f->val.Uint8);
			break;
		case TTMP_FT_Int16:
			fprintf(fp, "Int16, value %d\n", (int) f->val.Int16);
			break;
		case TTMP_FT_Uint16:
			fprintf(fp, "Uint16, value %u\n", (unsigned) f->val.Uint16);
			break;
		case TTMP_FT_Int32:
			fprintf(fp, "Int32, value %d\n", (int) f->val.Int32);
			break;
		case TTMP_FT_Uint32:
			fprintf(fp, "Uint32, value %u\n", (unsigned) f->val.Uint32);
			break;
		case TTMP_FT_Real32:
			fprintf(fp, "Real32, value %f\n", (double) f->val.Real32);
			break;
		case TTMP_FT_Real64:
			fprintf(fp, "Real64, value %f\n", f->val.Real64);
			break;
		case TTMP_FT_Real:
			fprintf(fp, "Real, value %f\n", f->val.Real);
			break;
		case TTMP_FT_Bool:
			fprintf(fp, "Bool, value %d\n", (int) f->val.Bool);
			break;
		case TTMP_FT_Chr:
			fprintf(fp, "Chr, value %u\n", (unsigned) f->val.Chr);
			break;
		case TTMP_FT_UniChr:
			fprintf(fp, "UniChr, value %u\n", (unsigned) f->val.UniChr);
			break;
		case TTMP_FT_Long64:
			fprintf(fp, "Long64, value %" PRINTF_LL "d\n", (longlong) f->val.Long64);
			break;
		case TTMP_FT_Str:
			fprintf(fp, "Str, size %u, value \"%s\"\n", f->size, (const char *) f->val.data);
			break;
		case TTMP_FT_UniStr:
			fprintf(fp, "UniStr, size %u, value:\n", f->size);
			hexdump(fp, f->val.data, f->size);
			break;
		case TTMP_FT_Binary:
			fprintf(fp, "Binary, size %u, value:\n", f->size);
			hexdump(fp, f->val.data, f->size);
			break;
		default:
			fprintf(fp, "Unknown type %d!\n", f->type);
			break;
		}
	}

	return TTMP_ERR_NONE;
}

int TTMPMessageMarshal(cTTMPMessage inmsg, TUint8 **outbuf, TUint32 *outsize)
{
	TUint32 i;
	TUint32 size;
	TTMPField *f;
	TUint8 len;
	TUint8 *buf;

	CHECK_OBJ(MSG, inmsg);
	if (outbuf == NULL || outsize == NULL)
		return TTMP_ERR_ARGUMENT;

	/* Message starts with a magic number, the total size, and the number of fields */
	size = TTMP_FS_Uint32 + TTMP_FS_Uint32 + TTMP_FS_Uint32;
	for (i = 0; i < inmsg->numfields; ++i) {
		/* Each field will be prefixed by:
		 * - a byte containing the length of the field name
		 * - the field name itself (no NUL terminator)
		 * - a type byte, from TTMP_FT_*
		 */
		f = &inmsg->fields[i];
		size += TTMP_FS_Uint8 + (TUint32) strlen(f->name) + TTMP_FS_Uint8;
		switch (f->type) {
		case TTMP_FT_Int8:
			size += TTMP_FS_Int8;
			break;
		case TTMP_FT_Uint8:
			size += TTMP_FS_Uint8;
			break;
		case TTMP_FT_Int16:
			size += TTMP_FS_Int16;
			break;
		case TTMP_FT_Uint16:
			size += TTMP_FS_Uint16;
			break;
		case TTMP_FT_Int32:
			size += TTMP_FS_Int32;
			break;
		case TTMP_FT_Uint32:
			size += TTMP_FS_Uint32;
			break;
		case TTMP_FT_Real32:
			size += TTMP_FS_Real32;
			break;
		case TTMP_FT_Real64:
			size += TTMP_FS_Real64;
			break;
		case TTMP_FT_Real:
			size += TTMP_FS_Real;
			break;
		case TTMP_FT_Bool:
			size += TTMP_FS_Bool;
			break;
		case TTMP_FT_Chr:
			size += TTMP_FS_Chr;
			break;
		case TTMP_FT_UniChr:
			size += TTMP_FS_UniChr;
			break;
		case TTMP_FT_Long64:
			size += TTMP_FS_Long64;
			break;
		case TTMP_FT_Str:
		case TTMP_FT_UniStr:
		case TTMP_FT_Binary:
			/* Strings and blobs are prefixed by their size */
			size += TTMP_FS_Uint32 + f->size;
			break;
		default:
			/* Shouldn't happen */
			TTMPERR("Unexpected field type %u\n", (unsigned) f->type);
			return TTMP_ERR_INTERNAL;
		}
	}
	/* Message ends with a CRC */
	size += TTMP_FS_Uint32;

	/* Allocate space for the marshal buffer */
	buf = malloc(size);
	if (buf == NULL)
		return TTMP_ERR_NOMEMORY;
#ifdef DEBUG
	/* To facilitate debugging, fill buffer with a known value */
	memset(buf, 0xC9, size);
#endif

	/* Set buffer result for caller */
	*outbuf = buf;
	*outsize = size;

	/* Define some dirty streaming macros */
#define STREAM(_type, _val)                                                   \
	do {                                                                      \
		T##_type tmp = (_val);                                                \
		memcpy(buf, &tmp, sizeof(T##_type));                                  \
		buf += TTMP_FS_##_type;                                               \
	} while (0)

#define STREAM_DATA(_data, _size)                                             \
	do {                                                                      \
		memcpy(buf, _data, _size);                                            \
		buf += (_size);                                                       \
	} while (0)

#define STREAM_TYPE(_type)                                                    \
	STREAM(Uint8, TTMP_FT_##_type)

#define STREAM_ELEM(_type)                                                    \
	do {                                                                      \
		STREAM_TYPE(_type);                                                   \
		STREAM(_type, f->val._type);                                          \
	} while (0)

#define STREAM_BLOB(_type)                                                    \
	do {                                                                      \
		STREAM_TYPE(_type);                                                   \
		STREAM(Uint32, f->size);                                              \
		STREAM_DATA(f->val.data, f->size);                                    \
	} while (0)

	/* Now actually stream message data into it */
	STREAM(Uint32, TTMP_MSG_MAGIC);
	STREAM(Uint32, size);
	STREAM(Uint32, inmsg->numfields);
	for (i = 0; i < inmsg->numfields; ++i) {
		f = &inmsg->fields[i];
		len = (TUint8) strlen(f->name);
		STREAM(Uint8, len);
		STREAM_DATA(f->name, len);
		switch (f->type) {
		case TTMP_FT_Int8:
			STREAM_ELEM(Int8);
			break;
		case TTMP_FT_Uint8:
			STREAM_ELEM(Uint8);
			break;
		case TTMP_FT_Int16:
			STREAM_ELEM(Int16);
			break;
		case TTMP_FT_Uint16:
			STREAM_ELEM(Uint16);
			break;
		case TTMP_FT_Int32:
			STREAM_ELEM(Int32);
			break;
		case TTMP_FT_Uint32:
			STREAM_ELEM(Uint32);
			break;
		case TTMP_FT_Real32:
			STREAM_ELEM(Real32);
			break;
		case TTMP_FT_Real64:
			STREAM_ELEM(Real64);
			break;
		case TTMP_FT_Real:
			STREAM_ELEM(Real);
			break;
		case TTMP_FT_Bool:
			STREAM_ELEM(Bool);
			break;
		case TTMP_FT_Chr:
			STREAM_ELEM(Chr);
			break;
		case TTMP_FT_UniChr:
			STREAM_ELEM(UniChr);
			break;
		case TTMP_FT_Long64:
			STREAM_ELEM(Long64);
			break;
		case TTMP_FT_Str:
			STREAM_BLOB(Str);
			break;
		case TTMP_FT_UniStr:
			STREAM_BLOB(UniStr);
			break;
		case TTMP_FT_Binary:
			STREAM_BLOB(Binary);
			break;
		}
	}
	/* Append CRC */
	STREAM(Uint32, TTMPCRC(*outbuf, size - TTMP_FS_Uint32));

	/* Undefine dirty macros to prevent mistakes later */
#undef STREAM
#undef STREAM_DATA
#undef STREAM_NAME
#undef STREAM_TYPE
#undef STREAM_ELEM
#undef STREAM_BLOB

	return TTMP_ERR_NONE;
}

int TTMPMessageUnmarshal(const TUint8 *inbuf, TUint32 insize, TTMPMessage *outmsg)
{
	const TUint8 *buf;
	TUint32 left;
	TUint32 magic;
	TUint32 size;
	TUint32 numfields;
	TTMPMessage msg;
	int res;
	TUint32 i;
	TUint8 len;
	TTMPFieldName name;
	TUint8 type;
	TTMPValue val;
	TUint32 crcbuf;
	TUint32 crccalc;

	if (inbuf == NULL || insize <= 0 || outmsg == NULL) {
		TTMPDBG("inbuf, insize or outmsg invalid\n");
		return TTMP_ERR_ARGUMENT;
	}

	/* Make sure the message is long enough */
	if (insize < TTMP_FS_Uint32 + TTMP_FS_Uint32 + TTMP_FS_Uint32) {
		TTMPDBG("message too small\n");
		return TTMP_ERR_MSGCORR;
	}

	/* Define some dirty unstreaming macros */
#define UNSTREAM(_type, _val)                                                 \
	do {                                                                      \
		memcpy(&(_val), buf, sizeof(T##_type));                               \
		buf += TTMP_FS_##_type;                                               \
	} while (0)

#define UNSTREAM_DATA(_data, _size)                                           \
	do {                                                                      \
		memcpy(_data, buf, _size);                                            \
		buf += (_size);                                                       \
	} while (0)

#define UNSTREAM_ELEM(_type)                                                  \
	do {                                                                      \
		UNSTREAM(_type, val._type);                                           \
		res = TTMPSetField##_type(msg, name, val._type);                      \
	} while (0)

#define UNSTREAM_BLOB(_type)                                                  \
	do {                                                                      \
		TUint32 size;                                                         \
		UNSTREAM(Uint32, size);                                               \
		res = TTMPFieldSetBlob(msg, name, TTMP_FT_##_type, buf, size);        \
		buf += size;                                                          \
	} while (0)

	buf = inbuf;
	left = insize;

	/* Check magic */
	UNSTREAM(Uint32, magic);
	if (magic != TTMP_MSG_MAGIC) {
		TTMPDBG("magic %#x doesn't match expected %#x\n", magic, TTMP_MSG_MAGIC);
		return TTMP_ERR_MSGCORR;
	}

	/* Read size and number of fields */
	UNSTREAM(Uint32, size);
	UNSTREAM(Uint32, numfields);
	if (size != insize) {
		TTMPDBG("message size %u doesn't match insize %u\n", size, insize);
		return TTMP_ERR_MSGCORR;
	}

	/* Size is okay now, allocate a message */
	res = TTMPMessageCreate(&msg);
	if (res != TTMP_ERR_NONE) {
		TTMPDBG("unable to allocate message\n");
		return res;
	}

	/* Now actually unstream data into the message */
	for (i = 0; i < numfields; ++i) {
		/* Unstream field name */
		UNSTREAM(Uint8, len);
		if (len == 0 || len > TTMP_MAX_FIELDNAME) {
			TTMPDBG("field %u len %u invalid\n", i, (unsigned) len);
			/* Don't leak in case of errors */
			TTMPMessageDestroy(msg);
			return TTMP_ERR_MSGCORR;
		}
		UNSTREAM_DATA(name, len);
		name[len] = '\0';
		/* Unstream field type */
		UNSTREAM(Uint8, type);
		switch (type) {
		case TTMP_FT_Int8:
			UNSTREAM_ELEM(Int8);
			break;
		case TTMP_FT_Uint8:
			UNSTREAM_ELEM(Uint8);
			break;
		case TTMP_FT_Int16:
			UNSTREAM_ELEM(Int16);
			break;
		case TTMP_FT_Uint16:
			UNSTREAM_ELEM(Uint16);
			break;
		case TTMP_FT_Int32:
			UNSTREAM_ELEM(Int32);
			break;
		case TTMP_FT_Uint32:
			UNSTREAM_ELEM(Uint32);
			break;
		case TTMP_FT_Real32:
			UNSTREAM_ELEM(Real32);
			break;
		case TTMP_FT_Real64:
			UNSTREAM_ELEM(Real64);
			break;
		case TTMP_FT_Real:
			UNSTREAM_ELEM(Real);
			break;
		case TTMP_FT_Bool:
			UNSTREAM_ELEM(Bool);
			break;
		case TTMP_FT_Chr:
			UNSTREAM_ELEM(Chr);
			break;
		case TTMP_FT_UniChr:
			UNSTREAM_ELEM(UniChr);
			break;
		case TTMP_FT_Long64:
			UNSTREAM_ELEM(Long64);
			break;
		case TTMP_FT_Str:
			UNSTREAM_BLOB(Str);
			break;
		case TTMP_FT_UniStr:
			UNSTREAM_BLOB(UniStr);
			break;
		case TTMP_FT_Binary:
			UNSTREAM_BLOB(Binary);
			break;
		}
	}
	/* Check CRC */
	UNSTREAM(Uint32, crcbuf);
	crccalc = TTMPCRC(inbuf, insize - TTMP_FS_Uint32);
	if (crcbuf != crccalc) {
		TTMPDBG("crc %#x doesn't match calculated %#x\n", crcbuf, crccalc);
		/* Don't leak in case of errors */
		TTMPMessageDestroy(msg);
		return TTMP_ERR_MSGCORR;
	}

	/* Set message result for caller */
	*outmsg = msg;

	return TTMP_ERR_NONE;
}

/* EOF */
