/* TTMPTestConn.c
 *
 * TTMPConnection test program.
 *
 * Copyright (C) 2005-2006 TomTom BV <http://www.tomtom.com/>
 * Author 0: Dimitry Andric <dimitry.andric@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2
 * as published by the Free Software Foundation.
 */

#include <stdio.h>
#include "TTMP.h"
#include "TTMPTestPlat.h"

#ifdef DMALLOC
#include "dmalloc.h"
#endif

static void print_error(const char *msg, int err)
{
	char buf[256];
	int res;

	res = TTMPErrorString(err, buf, sizeof buf);
	if (res != TTMP_ERR_NONE)
		printf("%s, error %d: Unable to get error string: error %d\n", msg, err, res);
	else
		printf("%s, error %d: %s\n", msg, err, buf);
}

int main(void)
{
	TTMPConnection conn;
	int res;
	TTMPMessage msg;

	signals_init();
	sockets_init();

	printf("Creating connection...\n");
	res = TTMPConnectionCreate(&conn, TTMP_TO_DEFAULT, VER_MAJOR, VER_MINOR);
	if (res != TTMP_ERR_NONE) {
		print_error("Unable to create connection", res);
		goto error_exit;
	}

	printf("Connecting to \"%s\"...\n", TEST_ADDRESS);
	res = TTMPConnectionConnect(conn, TEST_ADDRESS);
	if (res != TTMP_ERR_NONE) {
		print_error("Unable to connect", res);
		goto error_destroy_conn;
	}

	printf("Connected, creating message...\n");
	res = TTMPMessageCreate(&msg);
	if (res != TTMP_ERR_NONE) {
		print_error("Unable to create message", res);
		goto error_destroy_conn;
	}

	res = TTMPSetFieldInt16(msg, "command", 42);
	if (res != TTMP_ERR_NONE) {
		print_error("Unable to set command field", res);
		goto error_destroy_msg;
	}

	res = TTMPSetFieldReal(msg, "value", 3.14159265358979323846);
	if (res != TTMP_ERR_NONE) {
		print_error("Unable to set value field", res);
		goto error_destroy_msg;
	}
	
	res = TTMPSetFieldStr(msg, "description", "Feature creep is the root of all evil.");
	if (res != TTMP_ERR_NONE) {
		print_error("Unable to set description field", res);
		goto error_destroy_msg;
	}

	printf("Sending message 1...\n");
	res = TTMPConnectionSend(conn, msg);
	if (res != TTMP_ERR_NONE) {
		print_error("Unable to send message 1", res);
		goto error_destroy_msg;
	}

	printf("Clearing message...\n");
	res = TTMPMessageClear(msg);
	if (res != TTMP_ERR_NONE) {
		print_error("Unable to clear message", res);
		goto error_destroy_msg;
	}

	res = TTMPSetFieldLong64(msg, "foo", (TLong64) LONGLONG(1234567890123456789));
	if (res != TTMP_ERR_NONE) {
		print_error("Unable to set foo field", res);
		goto error_destroy_msg;
	}

	res = TTMPSetFieldStr(msg, "bar", "All Work And No Play Makes Jack A Dull Boy.\nAll Work And No Play Makes Jack A Dull Boy.\nAll Work And No Play Makes Jack A Dull Boy.\n");
	if (res != TTMP_ERR_NONE) {
		print_error("Unable to set bar field", res);
		goto error_destroy_msg;
	}

	printf("Sending message 2...\n");
	res = TTMPConnectionSend(conn, msg);
	if (res != TTMP_ERR_NONE) {
		print_error("Unable to send message 2", res);
		goto error_destroy_msg;
	}

	printf("Messages sent successfully, destroying it...\n");
	res = TTMPMessageDestroy(msg);
	if (res != TTMP_ERR_NONE) {
		print_error("Unable to destroy message", res);
		goto error_destroy_conn;
	}

	printf("Destroying connection...\n");
	res = TTMPConnectionDestroy(conn);
	if (res != TTMP_ERR_NONE) {
		print_error("Unable to destroy connection", res);
		goto error_exit;
	}

	printf("Done.\n");
	return 0;

error_destroy_msg:
	printf("Destroying message...\n");
	res = TTMPMessageDestroy(msg);
	if (res != TTMP_ERR_NONE) {
		print_error("Unable to destroy message", res);
	}

error_destroy_conn:
	printf("Destroying connection...\n");
	res = TTMPConnectionDestroy(conn);
	if (res != TTMP_ERR_NONE) {
		print_error("Unable to destroy connection", res);
	}

error_exit:
	return 1;
}

/* EOF */
