/* TTMPTestMsg.c
 *
 * TTMPMessage test program.
 *
 * Copyright (C) 2005-2006 TomTom BV <http://www.tomtom.com/>
 * Author 0: Dimitry Andric <dimitry.andric@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2
 * as published by the Free Software Foundation.
 */

//#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "TTMP.h"
#include "TTMPPlat.h"
#include "hexdump.h"

#define TEST_MARSHAL
#define TEST_UNMARSHAL

#ifdef TEST_MARSHAL
#include "TTMPMsg.h"
#endif

#ifdef DMALLOC
#include "dmalloc.h"
#endif

#define SMALL
#define SEED 1127229073

#if defined SMALL
#define NUM_MSGS   10
#define MAX_FIELDS 10
#define MAX_STR    128
#define MAX_UNISTR 128
#define MAX_BINARY 128
#elif defined MEDIUM
#define NUM_MSGS   100
#define MAX_FIELDS 100
#define MAX_STR    256
#define MAX_UNISTR 256
#define MAX_BINARY 256
#elif defined LARGE
#define NUM_MSGS   1000
#define MAX_FIELDS 1000
#define MAX_STR    1024
#define MAX_UNISTR 1024
#define MAX_BINARY 1024
#elif defined HUGE
#define NUM_MSGS   5000
#define MAX_FIELDS 5000
#define MAX_STR    8192
#define MAX_UNISTR 4096
#define MAX_BINARY 8192
#else
#error Would you like small, medium or large fries with that?
#endif

#ifdef DMALLOC
#define CHECK_HEAP do {	dmalloc_verify(NULL); dmalloc_log_stats(); } while (0)
#else
#define CHECK_HEAP do {} while (0)
#endif

static inline void randomize(void)
{
	unsigned seed;
#ifdef SEED
	seed = SEED;
#else
	struct timeval tv;
	gettimeofday(&tv, NULL);
	seed = tv.tv_sec ^ tv.tv_usec;
#endif
	printf("Random seed: %u\n", seed);
	srandom(seed);
}

static inline longlong random64(void)
{
	return ((longlong) random() << 32) | (longlong) random();
}

int main(void)
{
	TUint32 i;
	TTMPMessage msgs[NUM_MSGS];
	int res;
	TUint32 num;
	TUint32 j;
	TTMPFieldName name;
	TChr str[MAX_STR];
	TUint32 len;
	TUint32 k;
	TUniChr unistr[MAX_UNISTR];
	TUint8 binary[MAX_BINARY];
#ifdef TEST_MARSHAL
	TUint8 *buf;
	TUint32 size;
#ifdef TEST_UNMARSHAL
	TTMPMessage msg;
#endif
#endif

	randomize();

	printf("Creating messages...\n");
	for (i = 0; i < NUM_MSGS; ++i) {
		printf("Message %d\n", i);
		res = TTMPMessageCreate(&msgs[i]);
		if (res != TTMP_ERR_NONE) {
			printf("Unable to create message %d, error %d\n", i, res);
			return 1;
		}
		num = random() % MAX_FIELDS;
		for (j = 0; j < num; ++j) {
			snprintf(name, TTMP_MAX_FIELDNAME, "name%ld", random() % 100000);
			//snprintf(name, TTMP_MAX_FIELDNAME, "name%d", j);
			printf("%2d: name \"%s\", type ", j, name);
			switch (random() % 16) {
			case 0:
			{
				TInt8 val = (TInt8) random();
				printf("Int8, value %d\n", (int) val);
				res = TTMPSetFieldInt8(msgs[i], name, val);
				break;
			}
			case 1:
			{
				TUint8 val = (TUint8) random();
				printf("Uint8, value %u\n", (unsigned) val);
				res = TTMPSetFieldUint8(msgs[i], name, val);
				break;
			}
			case 2:
			{
				TInt16 val = (TInt16) random();
				printf("Int16, value %d\n", (int) val);
				res = TTMPSetFieldInt16(msgs[i], name, val);
				break;
			}
			case 3:
			{
				TUint16 val = (TUint16) random();
				printf("Uint16, value %u\n", (unsigned) val);
				res = TTMPSetFieldUint16(msgs[i], name, val);
				break;
			}
			case 4:
			{
				TInt32 val = (TInt32) random();
				printf("Int32, value %d\n", (int) val);
				res = TTMPSetFieldInt32(msgs[i], name, val);
				break;
			}
			case 5:
			{
				TUint32 val = (TUint32) random();
				printf("Uint32, value %u\n", (unsigned) val);
				res = TTMPSetFieldUint32(msgs[i], name, val);
				break;
			}
			case 6:
			{
				TReal32 val = random() / (TReal32) 100000.0;
				printf("Real32, value %f\n", (double) val);
				res = TTMPSetFieldReal32(msgs[i], name, val);
				break;
			}
			case 7:
			{
				TReal64 val = random() / (TReal64) 100000.0;
				printf("Real64, value %f\n", val);
				res = TTMPSetFieldReal64(msgs[i], name, val);
				break;
			}
			case 8:
			{
				TReal val = random() / (TReal) 100000.0;
				printf("Real, value %f\n", val);
				res = TTMPSetFieldReal(msgs[i], name, val);
				break;
			}
			case 9:
			{
				TBool val = (TBool) random();
				printf("Bool, value %d\n", (int) val);
				res = TTMPSetFieldBool(msgs[i], name, val);
				break;
			}
			case 10:
			{
				TChr val = (TChr) random();
				printf("Chr, value %u\n", (unsigned) val);
				res = TTMPSetFieldChr(msgs[i], name, val);
				break;
			}
			case 11:
			{
				TUniChr val = (TUniChr) random();
				printf("UniChr, value %u\n", (unsigned) val);
				res = TTMPSetFieldUniChr(msgs[i], name, val);
				break;
			}
			case 12:
			{
				TLong64 val = (TLong64) random64();
				printf("Long64, value %lld\n", (longlong) val);
				res = TTMPSetFieldLong64(msgs[i], name, val);
				break;
			}
			case 13:
			{
				snprintf(str, MAX_STR, "StringTest%ld", random() % 100000);
				printf("Str, value \"%s\"\n", str);
				res = TTMPSetFieldStr(msgs[i], name, str);
				break;
			}
			case 14:
			{
				snprintf(str, MAX_STR, "UnicodeTest%ld", random() % 100000);
				/* Make fake Unicode string */
				len = (TUint32) strlen(str) + 1;
				for (k = 0; k <= len; ++k)
					unistr[k] = (TUniChr) str[k];
				printf("UniStr, value:\n");
				hexdump(stdout, unistr, len * 2);
				res = TTMPSetFieldUniStr(msgs[i], name, unistr);
				break;
			}
			case 15:
			{
				len = random() % (MAX_BINARY - 1) + 1;
				for (k = 0; k < len; ++k)
					binary[k] = (TUint8) random();
				printf("Binary, value:\n");
				hexdump(stdout, binary, len);
				res = TTMPSetFieldBinary(msgs[i], name, binary, len);
				break;
			}
			}
			if (res != TTMP_ERR_NONE) {
				printf("Unable to set field %d, error %d\n", i, res);
				return 1;
			}
		}
		printf("Dumping message %d:\n", i);
		res = TTMPMessageDump(msgs[i], stdout);
		if (res != TTMP_ERR_NONE) {
			printf("Unable to dump message %d, error %d\n", i, res);
			return 1;
		}

#ifdef TEST_MARSHAL
		printf("Marshalling message %d...\n", i);
		res = TTMPMessageMarshal(msgs[i], &buf, &size);
		if (res != TTMP_ERR_NONE) {
			printf("Unable to marshal message %d, error %d\n", i, res);
			return 1;
		}
		printf("Marshalled message %d:\n", i);
		hexdump(stdout, buf, size);

#ifdef TEST_UNMARSHAL
		printf("Unmarshalling message %d...\n", i);
		res = TTMPMessageUnmarshal(buf, size, &msg);
		if (res != TTMP_ERR_NONE) {
			printf("Unable to unmarshal message %d, error %d\n", i, res);
			return 1;
		}
		printf("Unmarshalled message %d:\n", i);
		res = TTMPMessageDump(msgs[i], stdout);
		if (res != TTMP_ERR_NONE) {
			printf("Unable to dump temp message, error %d\n", res);
			return 1;
		}
		res = TTMPMessageDestroy(msg);
		if (res != TTMP_ERR_NONE) {
			printf("Unable to destroy message %d, error %d\n", i, res);
			return 1;
		}
#endif /* TEST_UNMARSHAL */
		free(buf);
#endif /* TEST_MARSHAL */
	}

	CHECK_HEAP;

	printf("\nDumping messages...\n");
	for (i = 0; i < NUM_MSGS; ++i) {
		printf("Message %d\n", i);
		res = TTMPMessageDump(msgs[i], stdout);
		if (res != TTMP_ERR_NONE) {
			printf("Unable to dump message %d, error %d\n", i, res);
			return 1;
		}
	}

	printf("\nDeleting messages...\n");
	for (i = 0; i < NUM_MSGS; ++i) {
		printf("Message %d\n", i);
		res = TTMPMessageDestroy(msgs[i]);
		if (res != TTMP_ERR_NONE) {
			printf("Unable to destroy message %d, error %d\n", i, res);
			return 1;
		}
	}

	CHECK_HEAP;

	return 0;
}

/* EOF */
